/*
 * gibber-xmpp-writer.h - Header for GibberXmppWriter
 * Copyright (C) 2006 Collabora Ltd.
 * @author Sjoerd Simons <sjoerd@luon.net>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef __GIBBER_XMPP_WRITER_H__
#define __GIBBER_XMPP_WRITER_H__

#include <glib-object.h>

#include "gibber-xmpp-stanza.h"

G_BEGIN_DECLS

typedef struct _GibberXmppWriter GibberXmppWriter;
typedef struct _GibberXmppWriterClass GibberXmppWriterClass;

struct _GibberXmppWriterClass {
    GObjectClass parent_class;
};

struct _GibberXmppWriter {
    GObject parent;
};

GType gibber_xmpp_writer_get_type (void);

/* TYPE MACROS */
#define GIBBER_TYPE_XMPP_WRITER \
  (gibber_xmpp_writer_get_type ())
#define GIBBER_XMPP_WRITER(obj) \
  (G_TYPE_CHECK_INSTANCE_CAST((obj), GIBBER_TYPE_XMPP_WRITER, \
  GibberXmppWriter))
#define GIBBER_XMPP_WRITER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_CAST((klass), GIBBER_TYPE_XMPP_WRITER, \
   GibberXmppWriterClass))
#define GIBBER_IS_XMPP_WRITER(obj) \
  (G_TYPE_CHECK_INSTANCE_TYPE((obj), GIBBER_TYPE_XMPP_WRITER))
#define GIBBER_IS_XMPP_WRITER_CLASS(klass) \
  (G_TYPE_CHECK_CLASS_TYPE((klass), GIBBER_TYPE_XMPP_WRITER))
#define GIBBER_XMPP_WRITER_GET_CLASS(obj) \
  (G_TYPE_INSTANCE_GET_CLASS ((obj), GIBBER_TYPE_XMPP_WRITER, \
   GibberXmppWriterClass))


GibberXmppWriter *gibber_xmpp_writer_new (void);
GibberXmppWriter *gibber_xmpp_writer_new_no_stream (void);

void gibber_xmpp_writer_stream_open (GibberXmppWriter *writer,
    const gchar *to, const gchar *from, const gchar *version,
    const guint8 **data, gsize *length);

void gibber_xmpp_writer_stream_close (GibberXmppWriter *writer,
    const guint8 **data, gsize *length);

gboolean gibber_xmpp_writer_write_stanza (GibberXmppWriter *writer,
    GibberXmppStanza *stanza, const guint8 **data, gsize *length,
    GError **error);

void gibber_xmpp_writer_flush (GibberXmppWriter *writer);

G_END_DECLS

#endif /* #ifndef __GIBBER_XMPP_WRITER_H__*/
