
#ifndef __GIBBER_NAMESPACES_H__
#define __GIBBER_NAMESPACES_H__

#define GIBBER_XMPP_NS_STREAM \
  "http://etherx.jabber.org/streams"

#define GIBBER_XMPP_NS_TLS \
  "urn:ietf:params:xml:ns:xmpp-tls"

#define GIBBER_XMPP_NS_SASL_AUTH \
  "urn:ietf:params:xml:ns:xmpp-sasl"

#define NS_DISCO_INFO \
  "http://jabber.org/protocol/disco#info"

#define NS_DISCO_ITEMS \
  "http://jabber.org/protocol/disco#items"

#define GIBBER_XMPP_NS_XHTML_IM \
  "http://jabber.org/protocol/xhtml-im"

#define GIBBER_XMPP_NS_IBB \
  "http://jabber.org/protocol/ibb"

#define GIBBER_XMPP_NS_AMP \
  "http://jabber.org/protocol/amp"

#define GIBBER_W3C_NS_XHTML \
  "http://www.w3.org/1999/xhtml"

#define GIBBER_TELEPATHY_NS_CAPS \
  "http://telepathy.freedesktop.org/caps"

#define GIBBER_TELEPATHY_NS_TUBES \
  "http://telepathy.freedesktop.org/xmpp/tubes"

#define GIBBER_TELEPATHY_NS_OLPC_ACTIVITY_PROPS \
  "http://laptop.org/xmpp/activity-properties"

#define GIBBER_XMPP_NS_SI \
  "http://jabber.org/protocol/si"

#define GIBBER_XMPP_NS_FEATURENEG \
  "http://jabber.org/protocol/feature-neg"

#define GIBBER_XMPP_NS_DATA \
  "jabber:x:data"

#define GIBBER_XMPP_NS_EVENT \
  "jabber:x:event"

#define GIBBER_XMPP_NS_STANZAS \
  "urn:ietf:params:xml:ns:xmpp-stanzas"

#define GIBBER_XMPP_NS_IQ_OOB \
  "jabber:iq:oob"

#define GIBBER_XMPP_NS_X_OOB \
  "jabber:x:oob"

#define GIBBER_XMPP_NS_JINGLE_ERRORS \
  "http://jabber.org/protocol/jingle#errors"

#define GIBBER_TELEPATHY_NS_CLIQUE \
  "http://telepathy.freedesktop.org/xmpp/clique"

#endif /* #ifndef __GIBBER_NAMESPACES_H__ */
