AMP = "http://jabber.org/protocol/amp"
DISCO_INFO = "http://jabber.org/protocol/disco#info"
DISCO_ITEMS = "http://jabber.org/protocol/disco#items"
IBB = "http://jabber.org/protocol/ibb"
IQ_OOB = "jabber:iq:oob"
MUC = 'http://jabber.org/protocol/muc'
MUC_OWNER = '%s#owner' % MUC
MUC_USER = '%s#user' % MUC
OLPC_ACTIVITIES = "http://laptop.org/xmpp/activities"
OLPC_ACTIVITIES_NOTIFY = "%s+notify" % OLPC_ACTIVITIES
OLPC_ACTIVITY = "http://laptop.org/xmpp/activity"
OLPC_ACTIVITY_PROPS = "http://laptop.org/xmpp/activity-properties"
OLPC_ACTIVITY_PROPS_NOTIFY = "%s+notify" % OLPC_ACTIVITY_PROPS
OLPC_BUDDY = "http://laptop.org/xmpp/buddy"
OLPC_BUDDY_PROPS = "http://laptop.org/xmpp/buddy-properties"
OLPC_BUDDY_PROPS_NOTIFY = "%s+notify" % OLPC_BUDDY_PROPS
OLPC_CURRENT_ACTIVITY = "http://laptop.org/xmpp/current-activity"
OLPC_CURRENT_ACTIVITY_NOTIFY = "%s+notify" % OLPC_CURRENT_ACTIVITY
PUBSUB = "http://jabber.org/protocol/pubsub"
SI = "http://jabber.org/protocol/si"
STANZA = "urn:ietf:params:xml:ns:xmpp-stanzas"
TELEPATHY_CAPS = 'http://telepathy.freedesktop.org/caps'
TUBES = "http://telepathy.freedesktop.org/xmpp/tubes"
X_OOB = "jabber:x:oob"
