#ifndef _SALUT_EXTENSIONS_H
#define _SALUT_EXTENSIONS_H

#include "extensions/_gen/svc.h"
#include "extensions/_gen/enums.h"
#include "extensions/_gen/gtypes.h"
#include "extensions/_gen/interfaces.h"

#endif /* _SALUT_EXTENSIONS_H */

